/<!-- ISCSI_MAN_PAGE -->/ {
    r udisks2-docs.xml.iscsi.man
    d }

/<!-- ISCSI_DBUS_INTERFACE -->/ {
    r udisks2-docs.xml.iscsi.dbus
    d }

/<!-- ISCSI_GENERATED_CODE -->/ {
    r udisks2-docs.xml.iscsi.generated
    d }
